import sys
from typing import Any

class Console:
    INFO_COLOR = '\033[94m'    # Blue
    WARNING_COLOR = '\033[93m' # Yellow
    ERROR_COLOR = '\033[91m'   # Red
    RESET_COLOR = '\033[0m'    # Reset to default
    BOLD = '\033[0m'           # Bold text
    UNDERLINE = '\033[4m'      # Underline text
    MAGENTA =  '\033[35m'

    def info(self, message: Any):
        self._print_message(message, self.INFO_COLOR)

    def write(self, message: Any):
        self._print_message(message, self.MAGENTA)
    def warning(self, message: str):
        self._print_message(message, self.WARNING_COLOR)

    def error(self, message: str):
        self._print_message(message, self.ERROR_COLOR)

    def _print_message(self, message: str, color: str):
        formatted_message = f"{color}{self.BOLD}{message}{self.RESET_COLOR}"
        if 'ipykernel' in sys.modules:
            from IPython.display import display, HTML
            display(HTML(f"<pre style='color:{color[2:-1]}; font-weight:bold;'>{message}</pre>"))
        else:
            print(formatted_message)


import time
import sys
from threading import Thread
from icloudservice.src.tools.format.color_text import AnsiColors

class ProgressIndicator:
    def __init__(self, message: str, interval: float = 0.2, color: str = AnsiColors.OKBLUE):
        self.message = message
        self.interval = interval
        self.color = color
        self.reset_color = AnsiColors.ENDC
        self.running = False
        self.thread = None
        self.start_time = None
        self.exception_occurred = False
        self.stopped = False
        self.spinner_chars = ['|', '/', '-', '\\','']
        self.colors = ['\033[91m', '\033[92m', '\033[93m', '\033[94m', '\033[95m', '\033[96m']
        self.reset_color = '\033[0m'

    def start(self):
        """Starts the progress indicator animation."""
        if self.thread is None:
            self.thread = Thread(target=self._run)
            self.running = True
            self.start_time = time.time()
            self.thread.start()

    def _run(self):
        """Internal logic for the progress indicator thread."""
        idx = 0
        color_idx = 0
        while self.running:
            for i in range(5):
                if not self.running:
                    break
                sys.stdout.write(f'\r{self.color}{self.message} {"." * (i + 1)}{self.reset_color}{self.colors[color_idx]}{self.spinner_chars[idx]}{self.reset_color}')
                sys.stdout.flush()
                idx = (idx + 1) % len(self.spinner_chars)
                color_idx = (color_idx + 1) % len(self.colors)
                time.sleep(self.interval)
        
        # Clear the final line
        sys.stdout.write('\r' + ' ' * (len(self.message) + 6) + '\r')
        sys.stdout.flush()

    def stop(self, final_message: str = None, final_message_color: str = AnsiColors.OKGREEN):
        """Stops the progress indicator animation and shows a final message."""
        self.running = False
        if self.thread:
            self.thread.join()
            self.thread = None
        # Clear the final line
        sys.stdout.write('\r' + ' ' * (len(self.message) + 4) + '\r')
        sys.stdout.flush()
        if final_message:
            duration = time.time() - self.start_time
            sys.stdout.write(f'{final_message_color}{final_message} (Duration: {duration:.2f} seconds){self.reset_color}\n')
            sys.stdout.flush()
        else:
            duration = time.time() - self.start_time
            sys.stdout.write(f'{final_message_color}(Duration: {duration:.2f} seconds){self.reset_color}\n')
            sys.stdout.flush()
        self.stopped = True
    def __enter__(self):
        """Starts the progress indicator animation when entering the context."""
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Stops the progress indicator animation when exiting the context."""
        if not self.stopped:
            if exc_type:
                self.exception_occurred = True
                self.stop(final_message=f"Exception: {exc_value}", final_message_color=AnsiColors.FAIL) 
            else:
                self.stop(final_message="Process completed", final_message_color=AnsiColors.OKGREEN)


import os


class FileUtils:
    @staticmethod
    def get_file_extension(key: str) -> str:
        """Get the file extension from the object key and determine the file type."""
        _, ext = os.path.splitext(key)
        ext = ext.lower()  # Convert to lowercase for consistency

        # Determine file type based on the extension
        if ext in ['.txt']:
            return 'Text File'
        elif ext in ['.json']:
            return 'JSON File'
        elif ext in ['.parquet']:
            return 'Parquet File'
        elif ext in ['.csv']:
            return 'CSV File'
        elif ext in ['.xml']:
            return 'XML File'
        else:
            return 'Unknown Type'
    @staticmethod
    def format_size(size_in_bytes: int) -> str:
        """Converts bytes to a more readable size format (KB, MB, GB)."""
        if size_in_bytes < 1024:
            return f"{size_in_bytes} B"
        elif size_in_bytes < 1024 ** 2:
            return f"{size_in_bytes / 1024:.2f} KB"
        elif size_in_bytes < 1024 ** 3:
            return f"{size_in_bytes / 1024 ** 2:.2f} MB"
        else:
            return f"{size_in_bytes / 1024 ** 3:.2f} GB"
        
    @staticmethod
    def get_file_name(file_path:str)->str:
        if not isinstance(file_path, str):
            return ''
        return os.path.basename(file_path)
    @staticmethod
    def get_pwd():
        return os.getcwd()
    
    @staticmethod
    def get_local_path_default(path:str):
        return os.path.join(FileUtils.get_pwd(), FileUtils.get_file_name(path))
    @staticmethod
    def get_path(path:str,name:str)->str:
        return os.path.join(path, name)


from icloudservice.src.tools.handler.table_data import TableData
from icloudservice.src.tools.format.table_render import TableRenderer

from typing import List, Dict, Any


class ITableData(TableData):
    def __init__(self, structure: List[Dict[str, Any]]):
        super().__init__(structure)
    
    def as_table(self) -> TableRenderer:
        return TableRenderer(self.data, self.columns)
    
    def __repr__(self) -> str:
        return self.as_table().render()
    


class AnsiColors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    ROW_COLOR = '\033[33m'
    HEADER_COLOR = '\033[96m'
    
    @staticmethod
    def color_text(text, color):
        return f"{color}{text}{AnsiColors.ENDC}"



from typing import List, Dict, Any, Union

class TableData:
    def __init__(self, data: List[Dict[str, Any]]):
        self.data = data
        if data:
            self.columns = list(data[0].keys())
        else:
            self.columns = []
    def get_column(self, column_name: str) -> List[Any]:
        return [row.get(column_name) for row in self.data]
    def get_row(self, index: int) -> Dict[str, Any]:
        if index < len(self.data):
            return self.data[index]
        else:
            raise IndexError("Row index out of range")
    def __getitem__(self, key: Union[int, str]) -> Union[Dict[str, Any], List[Any]]:
        if isinstance(key, int):
            return self.get_row(key)
        elif isinstance(key, str):
            return self.get_column(key)
        else:
            raise TypeError("Key must be an integer or string")
    def get_count(self) -> int:
        return len(self.data)
    
    def __repr__(self) -> str:
        return str(self.data)



from icloudservice.src.service.aws.aws_services import AWSService
from icloudservice.src.service.aws.aws_services import S3Service,LambdaService,EC2Service,SNSService,SQSService

class AWSCloud(AWSService):
    def __init__(self, access_key: str, secret_key: str,region:str = None):
        super().__init__(access_key,secret_key,region)
    def S3Service(self, bucket_name: str):
        return S3Service(self, bucket_name)

    def LambdaService(self):
        return LambdaService(self)

    def EC2Service(self):
        return EC2Service(self)

    def SQSService(self):
        return SQSService(self)

    def SNSService(self):
        return SNSService(self)


import platform
import importlib.metadata
from dataclasses import dataclass
from .src.provider.aws.aws_provider import AWSCloud
from .src.provider.azure.azure_provider import  AzureCloud
from .src.tools.print.console import Console
from .src.tools.format.color_text import AnsiColors as color
# Get the Python version
python_version = platform.python_version()

@dataclass
class Service():
    console = Console()

    def info(self):
        try:
            version = importlib.metadata.version('icloudservice')
        except importlib.metadata.PackageNotFoundError:
            version = 'unknown'
        # Get class name
        service_name = 'Cloud integration services for AWS and AZURE'
        log_message = (
        f"{color.HEADER}Version library:{color.WARNING}{version} \n"
        f"{color.HEADER}Name:{color.WARNING}icloudservice \n"
        f"{color.HEADER}Description:{color.WARNING}{service_name} \n"
        f"{color.HEADER}Python Version:{color.WARNING}{python_version} \n"
        )
        self.console.write(log_message)

@dataclass
class Provider():
    console = Console()

    def AWS(self,access_key:str,secret_key:str,region:str = None):
        """
        Initializes an AWS Cloud client with the provided access key and secret key.

        Args:
            access_key (str): AWS access key ID.
            secret_key (str): AWS secret access key.

        Returns:
            AWSCloud: An instance of the AWSCloud class.
        """
        return AWSCloud(access_key,secret_key,region)
    
    def AZURE(self,subscription_id):
        """
        Initializes an Azure Cloud client with the provided subscription ID.

        Args:
            subscription_id: Azure subscription ID.

        Returns:
            AzureCloud: An instance of the AzureCloud class.
        """
        return AzureCloud(subscription_id)


from functools import lru_cache
from typing import List, Dict, Any
from icloudservice.src.tools.format.color_text import AnsiColors as color

class TableRenderer:
    def __init__(self, data: List[Dict[str, Any]], columns: List[str]):
        self.data = data
        self.columns = columns
        

    def get_column_widths(self) -> List[int]:
        # Determine the maximum width for each column
        column_widths = [max(len(str(item)) for item in col) for col in zip(*self.data, self.columns)]
        column_widths = [max(width, len(header)) for width, header in zip(column_widths, self.columns)]
        return column_widths

    def render(self) -> str:
        if not self.data:
            return "No structure available"
        
        header_color = color.HEADER_COLOR
        row_color = color.ROW_COLOR
        reset_color = color.ENDC
        
        column_widths = self.get_column_widths()
        headers = self.columns
        rows = [list(row.values()) for row in self.data]
        
        # Calculate the maximum row width
        max_row_width = max(sum(len(str(item)) for item in row) + len(row) * 5 for row in rows)
        border_width = max(max_row_width, sum(column_widths) + len(headers) * 3 + 1)
        
        # Create the table as a string
        table_str = '+' + '-' * (border_width - 2) + '+\n'  # Top border
        
        # Header row
        table_str += f"|{header_color}{' | '.join(header.center(width) for header, width in zip(headers, column_widths))}{reset_color}|\n"
        table_str += '+' + '-' * (border_width - 2) + '+\n'  # Header separator
        
        # Data rows
        for row in rows:
            table_str += f"|{row_color}{' | '.join(str(item).center(width) for item, width in zip(row, column_widths))}{reset_color}|\n"
        
        table_str += '+' + '-' * (border_width - 2) + '+\n'  # Bottom border
        return table_str



import json
import botocore.session
import time
from icloudservice.src.tools.format.table_formatter import ITableData
from icloudservice.src.tools.progress.progress import ProgressIndicator
from icloudservice.src.tools.print.console import Console
from icloudservice.src.tools.util import  FileUtils as util
from icloudservice.src.tools.format.color_text import AnsiColors as color
class AWSService:
    def __init__(self, access_key: str, secret_key: str,region: str = None):
        self.access_key = access_key
        self.secret_key = secret_key
        self.region_name = region
        self.session = botocore.session.get_session()

    def create_client(self, service_name: str,region_name:str = None):
        return self.session.create_client(
            service_name,
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.secret_key,
            region_name=region_name or self.region_name
        )

class S3Service():
    def __init__(self, aws_service: AWSService, bucket_name: str,region: str = None):
        self.client = aws_service.create_client('s3',region_name=region)
        self.bucket_name = bucket_name
        self.console = Console()

    def upload(self, local_path: str, remote_path: str, use_local_filename: bool = True):
        local_file_name = util.get_file_name(local_path)
        # Check if remote_path is a directory and whether to use the local file name
        if remote_path.endswith('/') and use_local_filename:
            # If remote_path is a directory, append the local file name to it
            remote_path = util.get_path(remote_path, local_file_name)
        with ProgressIndicator(message="Uploading file") as progress:
            try:
                with open(local_path, 'rb') as file:
                    self.client.put_object(Bucket=self.bucket_name, Key=remote_path, Body=file)
            except Exception as e:
                progress.stop(final_message=f"Upload failed: {str(e)}", final_message_color=color.FAIL)

    def download(self, remote_path: str, local_path: str = None):
        if local_path is None:
            local_path = util.get_local_path_default(remote_path)
        with ProgressIndicator(message="Downloading file") as progress:
            try:
                response = self.client.get_object(Bucket = self.bucket_name, Key = remote_path)
                with open(local_path, 'wb') as file:
                    file.write(response['Body'].read())
            except Exception as e:
                progress.stop(final_message=f"Download failed: {str(e)}", final_message_color=color.FAIL)

    def list_files(self, directory: str,include_directories: bool = False) -> ITableData:
        """
        Lists objects in a specified directory of the S3 bucket and collects relevant details.

        Args:
            directory (str): The directory prefix to list objects from.

        Returns:
            List[Dict[str, Any]]: A list of dictionaries containing details of each object.
        """
        with ProgressIndicator(message="Waiting for listing objects...") as progress:
            progress.start()
            
            # Initialize the list to store object details
            s3list_objects = []

            # Retrieve the list of objects
            response = self.client.list_objects_v2(Bucket=self.bucket_name, Prefix=directory)
            
            # Check if 'Contents' is in the response
            objects = response.get('Contents', [])
            folders = set()
            directory_prefix_length = len(directory)
            # Process each object
            for obj in objects:
                key = obj.get('Key')
                if key.endswith('/') and include_directories and len(key) > directory_prefix_length:
                    if key[directory_prefix_length:].count('/') == 1:
                        folders.add(key)
                else:
                    # Add the object details to the list
                    if len(key) > directory_prefix_length and key[directory_prefix_length] != '/':
                        # Ensure that the object is directly in the directory and not in a subdirectory
                        # Get the position of the next '/' after the directory prefix
                        next_slash_index = key.find('/', directory_prefix_length)
                        if next_slash_index == -1:
                            
                            obj_details = {
                                'Key': util.get_file_name(key),
                                'Type':util.get_file_extension(key),
                                'Size': util.format_size(obj.get('Size')) if obj.get('Size') is not None else 'Unknown',
                                'LastModified': obj.get('LastModified')
                            }
                            s3list_objects.append(obj_details)
            
            #Add folder details to the list
            for folder in folders:
                folder_details = {
                    'Key': folder,
                    'Type':'folder',
                    'Size': None,                          # Folders don't have a size
                    'LastModified': None
                }
                s3list_objects.append(folder_details)
            
        return ITableData(s3list_objects)

    def list_buckets(self):
        response = self.client.list_buckets()
        return [bucket['Name'] for bucket in response.get('Buckets', [])]

class LambdaService:
    def __init__(self, aws_service: AWSService):
        self.client = aws_service.create_client('lambda')

    def invoke(self, function_name: str, payload: dict):
        response = self.client.invoke(FunctionName=function_name, Payload=json.dumps(payload))
        return response['Payload'].read().decode()

class EC2Service:
    def __init__(self, aws_service: AWSService):
        self.client = aws_service.create_client('ec2')
        self.console = Console()

    def start_instance(self, instance_id: str) -> ITableData:
        with ProgressIndicator(message=f"Waiting for instance to start {instance_id}") as progress:
            response = self.client.start_instances(InstanceIds=[instance_id])

            start_info = response['StartingInstances'][0]
            instance_id = start_info['InstanceId']
            current_state = start_info['CurrentState']['Name']
            previous_state = start_info['PreviousState']['Name']

            start_info_dict = []
            start_info_dict.append({
                "Instance ID": instance_id,
                "Current State": current_state,
                "Previous State": previous_state
            })

            #print(ITableData(start_info_dict))

            while True:
                instance_description = self.client.describe_instances(InstanceIds=[instance_id])
                instance = instance_description['Reservations'][0]['Instances'][0]
                instance_state = instance['State']['Name']

                if instance_state == 'running':
                    instance_dns = instance.get('PublicDnsName', 'No DNS available')
                    instance_name = self._get_instance_name(instance)
                    instance_details = []
                    instance_details.append({
                        "Instance ID": instance_id,
                        "State": instance_state,
                        "Name": instance_name,
                        "Os version": self.get_instance_os(instance_id),
                        "DNS": instance_dns
                    })
                    progress.stop(f"\nInstance {instance_id} is now fully running!")
                    return ITableData(instance_details)
                self.console.info(f"\nInstance {instance_id} is {instance_state}. Waiting for it to be fully running...")
                time.sleep(10)
    def _get_instance_name(self, instance):
        # Buscar el nombre en las etiquetas
        for tag in instance.get('Tags', []):
            if tag['Key'] == 'Name':
                return tag['Value']
        return 'No Name Tag'
    
    def stop_instance(self, instance_id: str) ->ITableData:
        with ProgressIndicator(message=f"Waiting for instance to stop {instance_id}") as progress:
            response = self.client.stop_instances(InstanceIds=[instance_id])
            instance_init = []

            for instance in response['StoppingInstances']:
                instance_data = {
                    "Instance ID": instance['InstanceId'],
                    "Current State": instance['CurrentState']['Name'],
                    "Previous State": instance['PreviousState']['Name']
                }
                instance_init.append(instance_data)
            instance_stop = []
            while True:
                instance_description = self.client.describe_instances(InstanceIds=[instance_id])
                instance = instance_description['Reservations'][0]['Instances'][0]
                instance_state = instance['State']['Name']

                if instance_state == 'stopped':
                    instance_stop.append({
                        "Instance ID": instance_id,
                        "State": instance_state,
                        "Name": self._get_instance_name(instance),
                        "Os version": self.get_instance_os(instance_id),
                        "DNS": instance.get('PublicDnsName', 'No DNS available')
                    })
                    break
                self.console.info(f"\nInstance {instance_id} is {instance_state}. Waiting for it to be fully stopped...")
                time.sleep(10)
            progress.stop(f'Instance {instance_id} is now fully stopped!')
        return ITableData(instance_stop)
                
    def describe_instances(self, instance_id: str = None) ->ITableData:
        with ProgressIndicator(message="Waiting ") as progress:
            # Fetch instances
            if instance_id:
                response = self.client.describe_instances(InstanceIds=[instance_id])
            else:
                response = self.client.describe_instances()
            
            
            instances = []
            # find all instances or the enter instance_ids 
            for reservation in response['Reservations']:
                for instance in reservation['Instances']:
                    instance_id = instance['InstanceId']
                    instance_state = instance['State']['Name']
                    instance_name = self._get_instance_name(instance)
                    instance_dns = instance.get('PublicDnsName', 'No DNS available')
                    instance_ip = instance.get('PublicIpAddress', 'No IP available')
                    os_info = self.get_instance_os(instance['InstanceId'])
                    instances.append({
                        "Instance ID": instance_id,
                        "Name": instance_name,
                        "Status":instance_state,
                        "Public DNS": instance_dns,
                        "Public IP": instance_ip,
                        "Type OS": os_info
                    })
            progress.stop('Finished successfully')
        return ITableData(instances)

    def get_instance_os(self, instance_id):
        response = self.client.describe_instances(InstanceIds=[instance_id])
        instance = response['Reservations'][0]['Instances'][0]
        ami_id = instance['ImageId']
        # Fetch AMI details
        ami_response = self.client.describe_images(ImageIds=[ami_id])
        ami = ami_response['Images'][0]
        # AMI description may contain OS information
        os_description = ami.get('Description', 'No OS information available')
        return os_description
    
    def list_instances(self) -> ITableData:
        return self.describe_instances()

class SQSService:
    def __init__(self, aws_service: AWSService):
        self.client = aws_service.create_client('sqs')

    def send_message(self, queue_url: str, message_body: str):
        self.client.send_message(QueueUrl=queue_url, MessageBody=message_body)

    def receive_message(self, queue_url: str):
        response = self.client.receive_message(QueueUrl=queue_url)
        return response.get('Messages', [])

class SNSService:
    def __init__(self, aws_service: AWSService):
        self.client = aws_service.create_client('sns')

    def publish_message(self, topic_arn: str, message: str):
        self.client.publish(TopicArn=topic_arn, Message=message)

    def list_topics(self):
        response = self.client.list_topics()
        return [topic['TopicArn'] for topic in response.get('Topics', [])]

class AWSCloud(AWSService):
    def __init__(self, access_key: str, secret_key: str):
        super().__init__(access_key, secret_key)

    def S3Service(self, bucket_name: str):
        return S3Service(self, bucket_name)

    def LambdaService(self):
        return LambdaService(self)

    def EC2Service(self):
        return EC2Service(self)

    def SQSService(self):
        return SQSService(self)

    def SNSService(self):
        return SNSService(self)
